#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb 24 23:44:05 2020

@author: carl
"""

import numpy as np

def normalize_spectra(method, y, wn=None, **kwargs):
    """
    Normalize the spectra in the matrix y (pixel,wavenum) with the given method
    """
    if method == 'mean':
        return (y.T / y.mean(axis=1)).T
    elif method == 'area':
         return (y.T / -np.trapz(y, wn, axis=1)).T
    elif method == 'wn':
        idx = (np.abs(wn - kwargs['wavenum'])).argmin()
        return (y.T / y[:, idx]).T
    elif method == 'max':
        return (y.T / y.max(axis=1)).T
    elif method == 'n2':
        return (y.T / np.sqrt((y * y).mean(axis=1))).T
    raise ValueError('undefined normalization method ' + method)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 14:01:08 2020

@author: carl
"""

import os.path
import numpy as np
import scipy.signal, scipy.io
from pymatreader import read_mat

from .opusreader import OpusReader

class SpectralData:
    """
    The list of current files and the raw data loaded from a hyperspectral image
    file in the preprocessing GUI.
    """
    def __init__(self):
        self.foldername = '' # Root dir
        self.filenames = []  # With full paths
        self.curFile = ''    # The currently loaded file
        self.wavenumber = None # array in order high->low
        self.wmin = 800
        self.wmax = 4000
        self.raw = np.empty((0,0)) # data in order (pixel, wavenumber)
        self.wh = (0, 0)  # Width and height in pixels
        self.image = None # Image loaded from raw data file; tuple (bytes, format_str)

    def set_width(self, w):
        try:
            w = int(w)
        except ValueError:
            w = 0
        if w <= 0:
            return False
        h = int(self.raw.shape[0] / w)
        if w * h != self.raw.shape[0]:
            return False
        self.wh = (w, h)
        return True

    def set_height(self, h):
        if self.set_width(h):
            self.wh = (self.wh[1], self.wh[0])
            return True
        return False

    def read_matrix(self, filename):
        """
        Read data from a file, with some error checking. The object is modified
        only if the file is successfully loaded.
        """
        wh = None
        image = None
        fext = os.path.splitext(filename)[1].lower()
#        opusformat = False
        if fext in ['.txt', '.csv', '.mat']:
            if fext == '.mat':
#                s = scipy.io.loadmat(filename)
                try:
                    s = read_mat(filename)
                except TypeError:
                    # Workaround for uint16_codec bug (pymatreader assumes mio5, not mio4)
                    s = scipy.io.loadmat(filename)
                # Assume data are in the biggest matrix in the file
                ss = max(s.items(), key=lambda k: np.size(k[1]) )[1]
                if 'wh' in s:
                    wh = s['wh'].flatten()
            else:
                ss = np.loadtxt(filename)

            if ss.ndim != 2 or ss.shape[0] < 10 or ss.shape[1] < 2:
                raise RuntimeError('file does not appear to describe an FTIR image matrix')
            d = -1 if ss[0,0] < ss[-1,0] else 1
            raw = ss[::d,1:].T
            wn = ss[::d,0]
        else:
            reader = OpusReader(filename)
            raw = reader.AB
            wn = reader.wavenum
            wh = reader.wh
            image = reader.image

        if (np.diff(wn) >= 0).any():
            raise RuntimeError('wavenumbers must be sorted')
        npix = raw.shape[0]
        if wh is not None:
            if len(wh) != 2:
                raise RuntimeError('Image size in "wh" must have length 2')
            wh = (int(wh[0]), int(wh[1]))
            if wh[0] * wh[1] != npix:
                raise RuntimeError('Image size in "wh" does not match data size')
            self.wh = wh
        elif npix != self.wh[0] * self.wh[1]:
            res = int(np.sqrt(npix))
            if npix == res * res:
                self.wh = (res, res)
            else:
                self.wh = (npix, 1)
        self.raw = raw
        self.wavenumber = wn
        self.wmin = self.wavenumber.min()
        self.wmax = self.wavenumber.max()
        self.image = image
        self.curFile = filename




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Carl Troein

Atmospheric and scattering correction

"""

import gc
from time import monotonic
import numpy as np
import sklearn.linear_model
import sklearn.cluster
#import statsmodels.multivariate.pca
from scipy.interpolate import PchipInterpolator
from scipy.signal import hilbert
#from scipy.io import loadmat, savemat

from .util import pca_nipals, find_wn_ranges

def kkre(wn, ref):
    wn2 = wn ** 2.
    wa = wn * ref
    kk = np.empty_like(wn)
    for i in range(len(wn)):
        with np.errstate(divide='ignore', invalid='ignore'):
            fg = wa / (wn2 - wn[i] ** 2.)
        if i == 0 or i == len(wn) - 1:
            fg[i] = 0
        else:
            fg[i] = (fg[i-1] + fg[i+1]) / 2
        kk[i] = 2/np.pi * np.trapz(x=wn, y=fg)
    if wn[0] < wn[-1]:
        return kk
    return -kk

def hilbert_n(wn, ref, zeropad=500):
    """
    Compute the Kramers-Kronig relations by Hilbert transform, extending the absorption spectrum
    with 0 to either size and resampling the spectrum at evenly spaced intervals if it is found to
    be unevenly sampled.
    """
    # Cache some data structures to avoid having to reinitialize them on every call.
    if not hasattr(hilbert_n, "wn") or hilbert_n.wn is not wn or hilbert_n.zp != zeropad:
        hilbert_n.wn = wn
        even = (wn[-1] - wn[0]) / (len(wn) - 1)
        diff = np.abs((np.diff(wn) - even) / even).mean()
        hilbert_n.zp = zeropad
        hilbert_n.evenspaced = diff < 1e-3
        hilbert_n.increasing = wn[0] < wn[-1]
#        print('hilbert',hilbert_n.evenspaced,hilbert_n.increasing)
        if not hilbert_n.evenspaced:
            hilbert_n.lin = np.linspace(min(wn[0], wn[-1]), max(wn[0], wn[-1]), len(wn))
        hilbert_n.npad = int(len(wn) / abs(wn[-1] - wn[0]) * zeropad)
        hilbert_n.nim = np.zeros((len(wn) + 2 * hilbert_n.npad))

    if hilbert_n.evenspaced:
        if hilbert_n.npad == 0:
            if hilbert_n.increasing:
                hilbert_n.nim = ref
            else:
                hilbert_n.nim = ref[::-1]
        elif hilbert_n.increasing:
            hilbert_n.nim[hilbert_n.npad:hilbert_n.npad+len(wn)] = ref
        else:
            hilbert_n.nim[hilbert_n.npad:hilbert_n.npad+len(wn)] = ref[::-1]
    else:
        if hilbert_n.increasing:
            hilbert_n.nim[hilbert_n.npad:hilbert_n.npad+len(wn)] = PchipInterpolator(wn, ref)(hilbert_n.lin)
        else:
            hilbert_n.nim[hilbert_n.npad:hilbert_n.npad+len(wn)] = PchipInterpolator(wn[::-1], ref[::-1])(hilbert_n.lin)
    nreal = -np.imag(hilbert(hilbert_n.nim))
    if hilbert_n.npad:
        nreal = nreal[hilbert_n.npad:-hilbert_n.npad]
    if hilbert_n.evenspaced:
        return nreal if hilbert_n.increasing else nreal[::-1]
    return PchipInterpolator(hilbert_n.lin, nreal)(wn)

def compute_model(wn, ref, n_components, a, d, bvals, konevskikh=True, linearcomponent=False,
                  variancelimit = None):
    """
    Compute the extinction matrix for resonant Mie scattering, then PCA transform it.
    Parameters:
    wn: array of wavenumbers
    ref: reference spectrum
    n_components: number of PCA components to use
    a: array of values for the parameter a (index of refraction)
    d: array of values for the parameter d*4pi (sphere size)
    bvals: number of values for parameter b (mixing of a and real part of n from absorption ref)
    konevskikh: if True, use the faster method by Konevskikh et al.
    variancelimit: if a number (around 0.9996), use as many PCA components as needed to
       explain this fraction of the variance of the extinction matrix
    """

    # Compute the scaled real part of the refractive index by Kramers-Kronig transform:
    # We skip a factor 2pi because it's normalized away anyway.
#    n_im = ref / wn
#    nkk = -np.imag(hilbert(n_im)) if wn[0] < wn[-1] else -np.imag(hilbert(n_im[::-1]))[::-1]

    # I'm a bit confused about the division/multiplication by wn.
    # Bassan's matlab code uses wn*ref in kkre
        # bassan2010 says k (=n_im) is proportional to ref
        # but various sources say k is ref/wn * const
    # My kkre reproduces's Bassan with the same wn*ref
    # My hilbert_n gives the same output with Hilbert transform of ref
    # Causin's python code Hilbert transforms ref for Bassan's algorith but ref/wn for Konevskikh's!
    # Solheim's matlab code Hilbert transforms ref/wn

    if konevskikh:
#        nim = ref / ref.max() / (wn * 100)
        nim = ref / (wn * 100)
        nre = hilbert_n(wn, nim, 300)
        nmin = nre.min()
#        print('refmax', ref.max(), 'nrange', nmin, nre.max())
        if nmin < -1:
            nre = nre / -nmin
            nim = nim / -nmin
        # My revised distribution of alpha_0 and gamma
#        alpha_0 = 1e-2 * np.linspace(d[0] * (a[0]-1), d[-1] * (a[-1]-1), len(a))
#        gamma = .25 * 2 * np.log(10) / np.pi * np.linspace(1 / alpha_0[0], 1 / alpha_0[-1], len(alpha_0))
        # Solheim's distributions of alpha_0 and gamma
        alpha_0 = 1e-2 * d * (a - 1)
        gamma = .25  * 2 * np.log(10) / np.pi / alpha_0
        Q = np.empty((len(alpha_0) * len(gamma), len(wn)))  # Initialize the extinction matrix

#        print('alpha_0', alpha_0)
#        print('gamma', gamma)
        # Build the extinction matrix
        n_row = 0
        for a0 in alpha_0:
            for g in gamma:
                rho = a0 * (1. + g * nre) * wn * 100
                denom = 1. / g + nre
                tanbeta = nim / denom
                beta = np.arctan2(nim, denom)
                cosb = np.cos(beta)
                cosbrho = cosb / rho

                # Following Konevskikh et al 2016
                Q[n_row] = 2. - 4. * cosbrho * (np.exp(-rho * tanbeta) *
                     (np.sin(rho - beta) + cosbrho * np.cos(rho - 2 * beta)) -
                     cosbrho * np.cos(2 * beta))

                n_row += 1
#        savemat('everything-p.mat', locals())
    else:
        nkk = kkre(wn, ref/wn) # should divide by wn here (or not multiply by wn in the function)
#        nkk = hilbert_n(wn, ref / wn, 300) # should divide by wn
        nkk = nkk / abs(nkk.min())

        # Build the extinction matrix
        Q = np.empty((len(a) * bvals * len(d), len(wn)))  # Initialize the extinction matrix
        n_row = 0
        for i in range(len(a)):
            b = np.linspace(0.0, a[i] - 1.01, bvals)  # Range of amplification factors of nkk
            for j in range(len(b)):
                n = a[i] + b[j] * nkk  # Compute the real refractive index
                for k in range(len(d)):
                    rho = d[k] * (n - 1.) * wn
                    #  Compute the extinction coefficients for each combination of a, b and d:
                    Q[n_row] = 2. - 4. / rho * np.sin(rho) + \
                               4. / (rho * rho) * (1. - np.cos(rho))
                    n_row += 1

    n_nonpca = 3 if linearcomponent else 2

#    savemat('everything_python.mat', {'Q':Q})

    # Orthogonalization of the model to improve numeric stability
    refn = ref / np.sqrt(ref@ref)
    Q = Q - (Q @ refn)[:,None] @ refn[None,:]

    # Perform PCA of the extinction matrix
    pca = pca_nipals(Q, ncomp=n_components, tol=1e-5, copy=False, explainedvariance=variancelimit)
    model = np.empty((n_nonpca + pca.shape[0], pca.shape[1]))
    model[n_nonpca:, :] = pca

#    # This method is up to 50% slower but gives additional info
#    pca = statsmodels.multivariate.pca.PCA(Q, ncomp=n_components,
#           method='nipals', tol=1e-5, demean=False, standardize=False)
#    esum = pca.eigenvals.cumsum()
#    n_components = np.min((np.searchsorted(esum, esum[-1] * variancelimit) + 1, len(esum)))
#    model = np.zeros((n_nonpca + n_components, len(wn)))
#    model[n_nonpca:, :] = pca.loadings[:,:n_components].T

    model[0,:] = ref
    model[1,:] = 1
    if linearcomponent:
        model[2,:] = np.linspace(0., 1., len(wn))
#        for i in range(1, n_nonpca):
#            model[i,:] = model[i,:] - refn * (model[i,:] @ refn)
#        for i in range(1, n_nonpca):
#            w = model[i, :] - np.sum(np.dot(model[i, :], b) * b for b in model[0:i, :])
#            model[i,:] = w / np.sqrt(w @ w)

#    savemat('everything_after-p.mat', locals())

    # Orthogonalization of the model to improve numeric stability (doing it after PCA is only
    # marginally slower)
#    for i in range(len(model)):
#        v = model[i, :]
#        w = v - np.sum(np.dot(v, b) * b for b in model[0:i, :])
#        model[i,:] = w / np.linalg.norm(w)

#    print('model',ref.sum(), model.sum())
    return model

def stable_rmiesc_clusters(iters, clusters):
    """
    Make a cluster size scheme for reliable convergence in rmiesc.
    Parameters:
    iters: The number of basic iterations to be used, preferably at least about 12-20
    Returns:
    array of cluster sizes (or zeros) for each iteration; this will be bigger than
    the input iterations
    """
    iters = max(2, iters) * 2
    cc = np.zeros(iters, dtype=np.int)
    cc[:iters//3] = 1
    cc[iters//2:iters*3//4] = clusters
#    cc = np.zeros(max(2, iters) * 3 // 2, dtype=np.int)
#    cc[:iters] = clusters
    return cc

def rmiesc(wn, app, ref, n_components=7, iterations=10, clusters=None,
           pcavariancelimit=None,
           verbose=False, a=np.linspace(1.1, 1.5, 10), d=np.linspace(2.0, 8.0, 10),
           bvals=10, progressCallback = None, progressPlotCallback=None,
           konevskikh=False, linearcomponent=True, weighted=False, renormalize=False,
           autoiterations=False, targetrelresiduals=0.95):
    """
    Correct scattered spectra using Bassan's algorithm. This implementation does no orthogonalization
    of the extinction matrix or PCA components relative to the reference, nor is the reference smoothed
    or filtered through a sum of gaussians as in the original Matlab implementation.
    Parameters:
    wn: sorted array of wavenumbers (high-to-low or low-to-high)
    app: apparent spectrum, shape (pixels, wavenumbers)
    ref: reference spectrum; array (wavenumbers)
    n_components: number of principal components to be calculated for the extinction matrix
    iterations: number of iterations of the algorithm
    clusters: if not None, cluster pixels into this many clusters in each iteration and use
        a common reference spectrum for each cluster. May be given as a list with one value per
        iteration, in which case 0 means to reuse clusters from the previous iteration and mix
        new/old references for stable convergence.
        If clusters is negative, use stable_rmiesc_clusters to generate the list.
    verbose: print progress information
    a: indexes of refraction to use in model
    d: sphere sizes to use in model, in micrometers
    bvals: number of values for the model parameter b
    progressCallback(int a, int b): callback function called to indicated that the processing
        is complete to a fraction a/b.
    progressPlotCallback(array ref, (a, b)): like the above with a (cluster) reference spectrum
    konevskikh: if True, use the faster method by Konevskikh et al.
    linearcomponent: if True, include a linear term in the model (used in Bassan's paper only).
    weighted: if true, downweight the 1800-2800 region when fitting the model.
    renormalize: if True, renormalize spectra against reference in each generation.
    autoiterations; if True, iterate until residuals stop improving
    targetrelresiduals: if autoiterations, stop when this relative change in residuals is seen
    Return: corrected apparent spectra (the best encountered if autoiterations, else the final ones)
    """

    # Make a rescaled copy of d and include the factor 4*pi
    d = d * 4e-4  * np.pi;

    # The input can be a single spectrum or a matrix of spectra. If the former, squeeze at the end.
    squeeze = False
    if app.ndim == 1:
        app = app[None,:]
        squeeze = True

    if weighted:
        weights = np.ones_like(wn)
        weights[range(*find_wn_ranges(wn, [[1800, 2800]])[0])] = .001 ** .5
        weights = weights[:, None]
    else:
        weights = None

    if np.isscalar(clusters):
        if clusters == 0:
            clusters = None
        elif clusters < 0:
            clusters = stable_rmiesc_clusters(iterations, -clusters)
            iterations = len(clusters)
        else:
            clusters = np.repeat(clusters, iterations)
    elif clusters is not None:
        if len(clusters) != iterations:
            raise ValueError('len(clusters) must match iterations')

    if progressCallback:
        # Compute the number of progress steps
        progressA = 0
        if clusters is None:
            progressB = 1 + (iterations > 1) * len(app)
        else:
            progressB = 1
            prev = 1
            for cl in clusters[1:]:
                if cl > 0:
                    prev = cl
                progressB += prev
    startt = monotonic()

    corrected = None # Just to get rid of warnings in the editor; will be set on iteration 0

    # Set parameters for automatic iteration control
    autoupadd = 1 # Residual going up counts as residual going down too little this many times
    automax = 5   # Stop when residual has gone down too little this many times

    if clusters is not None:
        # Cluster mode: In each iteration, after correcting all the spectra, cluster them.
        # Then take the mean of the corrected spectra in each cluster as the new reference
        # for that cluster in the next iteration. Spectra should still have their own reference
        # in the EMSC step. To avoid having to loop over all the spectra in a cluster when
        # solving that equation system, we remove the projection of the reference from the
        # apparent spectra.

        if progressPlotCallback:
            progressPlotCallback(ref, (0, iterations))
        projs = ((app @ ref) / (ref @ ref))[:,None] * ref
        app_deref = app - projs
        model = compute_model(wn, ref, n_components, a, d, bvals,
                              konevskikh=konevskikh, linearcomponent=linearcomponent,
                              variancelimit=pcavariancelimit)
        if weights is None:
            cons = np.linalg.lstsq(model.T, app_deref.T, rcond=None)[0]
        else:
            cons = np.linalg.lstsq(model.T * weights, app_deref.T * weights, rcond=None)[0]
        corrected = app - cons[1:].T @ model[1:]
        residuals = ((corrected - projs)**2).sum(1) #We compare to the previous correction, not the reference
        ref = []    # Force init
        if autoiterations:
            unimproved = np.zeros(len(app), dtype=int)

        if progressCallback:
            progstep = 1 # Current progress bar step size
            progressA += progstep
            progressCallback(progressA, progressB)

        for iteration in range(1, iterations):
            corrected[corrected < 0] = 0

            curc = clusters[iteration] # Current cluster size setting
            if curc > 0:
                progstep = curc

            # Possibly recluster the spectra and compute reference spectra
            if curc > 0:
                # Find the remaining spectra to process
                if autoiterations:
                    if curc != clusters[iteration-1]:
                        unimproved = np.zeros(len(app), dtype=int)
                        nds = len(app)
                        notdone = range(nds)
                    else:
                        notdone = np.where(unimproved <= automax)[0]
                        nds = len(notdone)
                    # Skip this iteration if every spectrum has stopped improving
                    if nds == 0:
                        progressA += progstep
                        if progressCallback:
                            progressCallback(progressA, progressB)
                        continue
                else:
                    nds = len(app)
                    notdone = range(nds)

                # Find the clusters of spectra
                curc = min(curc, nds)
                if curc == nds:
                    cindices = [np.array([i]) for i in notdone] # Each index in its own little array
                elif curc > 1:
                    kmeans = sklearn.cluster.MiniBatchKMeans(curc)
                    labels = np.zeros(len(app)) - 1
                    labels[notdone] = kmeans.fit_predict(corrected[notdone,:])
                    cindices = [ np.where(labels == i)[0] for i in range(curc) ] # Quadratic but simple
                else:
                    cindices = [ np.array(notdone) ] # One big cluster

                # Compute cluster references
                if(len(ref) != curc):
                    ref = np.zeros((curc, len(wn)))
                empties = 0
                for cl in range(curc):
                    if len(cindices[cl]):
                        ref[cl,:] = corrected[cindices[cl]].mean(0)
                    else:
                        empties += 1
                if empties:
                    print('Info: %d empty cluster(s) at iteration %d' % (empties, iteration))
            else:
                # Mix old reference and corrected spectrum. This requires the clusters
                # to remain unchanged.
                for cl in range(len(ref)):
                    if len(cindices[cl]):
                        # Stop processing clusters only when all members are done
                        if autoiterations and (unimproved[cindices[cl]] > automax).all():
                            cindices[cl] = []
                        else:
                            ref[cl,:] = .5 * corrected[cindices[cl]].mean(0) + .5 * ref[cl,:]

            w = (app * corrected).sum(1) / (corrected * corrected).sum(1)
            projs = w[:,None] * corrected
            app_deref = app - projs

            if progressPlotCallback:
                progressPlotCallback(ref, (iteration, iterations))

            for cl in range(len(ref)):
                ix = cindices[cl] # Indexes of spectra in this cluster
                if len(ix) > 0 and autoiterations:
                    ix = ix[unimproved[ix] <= automax]
                if len(ix) == 0:
                    continue

                model = compute_model(wn, ref[cl], n_components, a, d, bvals,
                                      konevskikh=konevskikh, linearcomponent=linearcomponent,
                                      variancelimit=pcavariancelimit)
                # Including the reference in the model gives results in agreement with the
                # unclustered approach.
                if weights is None:
                    cons = np.linalg.lstsq(model.T, app_deref[ix].T, rcond=None)[0]
                else:
                    cons = np.linalg.lstsq(model.T * weights, app_deref[ix].T * weights, rcond=None)[0]
                corrs = app[ix] - cons[1:].T @ model[1:]

                # We compare to the previous correction, not the reference
                resids = ((corrs - projs[ix])**2).sum(1)
                improved = resids < residuals[ix]

                if renormalize:
                    corrs = corrs / (1 + cons[0,:,None])

                if autoiterations:
                    iximp = ix[improved]  # Indexes of improved spectra
                    impmore = resids[improved] < residuals[iximp] * targetrelresiduals
                    unimproved[iximp[impmore]] = 0
                    unimproved[iximp[np.logical_not(impmore)]] += 1
                    unimproved[ix[np.logical_not(improved)]] += autoupadd

                    corrected[iximp, :] = corrs[improved, :]
                    residuals[iximp] = resids[improved]
#                    corrected[ix,:] = corrs
#                    residuals[ix] = resids
                else:
                    corrected[ix,:] = corrs
                    residuals[ix] = resids

                nimprov = improved.sum()

                if verbose:
                    print("iter %3d, cluster %3d (%5d px): avgres %7.3g  imprvd %4d  time %f" %
                          (iteration, cl, len(ix), resids.mean(), nimprov, monotonic()-startt))
                if progressCallback:
                    progressCallback(progressA + cl + 1, progressB)
            if progressCallback:
                progressA += progstep
                progressCallback(progressA, progressB)
            gc.collect()    # Because my old laptop was unhappy with RAM usage otherwise

    else:
        if renormalize:
            # More aggressive stopping if residual rises
            autoupadd = 3  # Residual going up counts as residual going down too little this many times
            automax = 3    # Stop when residual has gone down too little this many times
        ref = ref.copy()
        ref[ref < 0] = 0
        # For efficiency, compute the model from the input reference spectrum only once
        model = compute_model(wn, ref, n_components, a, d, bvals, konevskikh=konevskikh,
                              linearcomponent=linearcomponent, variancelimit=pcavariancelimit)

        if weights is None:
            cons = np.linalg.lstsq(model.T, app.T, rcond=None)[0]
        else:
            cons = np.linalg.lstsq(model.T * weights, app.T * weights, rcond=None)[0]
        corrected = app - cons[1:, :].T @ model[1:, :]
        if renormalize:
            corrected = corrected / cons[0, :, None]
        if autoiterations:
            residuals = ((corrected - model[0, :])**2).sum(1)

        if progressPlotCallback:
            progressPlotCallback(ref, (0, len(app) + 1))

        if verbose:
            print("all pixels, iter %2d  time %f" % (0, monotonic()-startt))
        if progressCallback:
            progressA += 1
            progressCallback(progressA, progressB)
        if iterations > 1:
            for s in range(len(app)):
                gc.collect()
                unimproved = 0
                ref = corrected[s, :]  # Corrected spectrum as new reference
                for iteration in range(1, iterations):
                    ref[ref < 0] = 0. # No negative values in reference spectrum
                    model = compute_model(wn, ref, n_components, a, d, bvals,
                                          konevskikh=konevskikh, linearcomponent=linearcomponent,
                                          variancelimit=pcavariancelimit)
                    if weights is None:
                        cons = np.linalg.lstsq(model.T, app[s], rcond=None)[0]
                    else:
                        cons = np.linalg.lstsq(model.T * weights, app[s] * weights[:, 0], rcond=None)[0]
                    corr = app[s] - cons[1:] @ model[1:, :]
                    if renormalize:
                        corr = corr / cons[0]
                    if autoiterations:
                        residual = ((corr - model[0, :])**2).sum()
                        if residual < residuals[s]:
                            corrected[s, :] = corr
                            unimproved =  unimproved + 1 if residual > residuals[s] * targetrelresiduals else 0
                            residuals[s] = residual
                        else:
                            unimproved += autoupadd
                        if unimproved > automax:
                            break
                    ref = corr
                if not autoiterations:
                    corrected[s, :] = corr
                    residual = ((corr / cons[0] - model[0, :])**2).sum()

                if verbose:
                    print("pixel %5d: iter %3d  residual %7.3g  time %f" %
                          (s, iteration, residual, monotonic()-startt))
                if progressCallback:
                    progressA += 1
                    progressCallback(progressA, progressB)
                if progressPlotCallback and len(app) < 50:
                    progressPlotCallback(ref, (s + 1, len(app) + 1))

    return corrected.squeeze() if squeeze else corrected





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 11 16:00:35 2020

@author: carl
"""

from PyQt5.QtWidgets import QSizePolicy
#from PyQt5.QtCore import pyqtSignal, pyqtSlot
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import matplotlib.pyplot as plt

class WhiteLightWidget(FigureCanvas):
    def __init__(self, parent=None):
        self.fig = Figure()
        self.ax = self.fig.add_subplot(111)
        self.ax.set_axis_off()
        self.ax.set_position([0,0,1,1])
        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)
        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)
        self.img = None

    def load(self, filename, format=None):
        self.ax.clear()
        self.ax.set_axis_off()
        self.img = None
        if filename is not None:
            try:
                self.img = plt.imread(filename, format=format)
                self.ax.imshow(self.img)
            except FileNotFoundError:
                pass
        self.draw_idle()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 27 22:55:02 2019

@author: carl
"""

import json
import traceback
import os.path
from collections import namedtuple
import numpy as np
import scipy.signal, scipy.io
from PyQt5.QtCore import QObject, pyqtSignal, pyqtSlot

from octavvs.io import SpectralData
from octavvs.algorithms import baseline, correction, normalization

class PrepParameters:
    """
    A class representing all the settings that can be made in the preprocessing UI,
    saved/loaded and used to start a batch job. The default values are only used if
    loading a file where some values are missing.
    """
    def __init__(self):
        self.fileFilter = '*.mat'
        self.saveExt = '_prep'
        self.plotMethod = 0
        self.plotColors = ''
        self.plotWavenum = 0
        self.spectraCount = 0
        self.spectraAuto = False
        self.acDo = False
        self.acSpline = True
        self.acLocal = True
        self.acSmooth = True
        self.acReference = ''
        self.scDo = False
        self.scRef = 'Casein'
        self.scOtherRef = ''
        self.scIters = 50
        self.scClusters = 30
        self.scStable = True
        self.scAlgorithm = 'Bassan'
        self.scResolution = 10
        self.scAmin = 1.1
        self.scAmax = 1.4
        self.scDmin = 2
        self.scDmax = 8
        self.scLinear = True
        self.scRenormalize = False
        self.scPCADynamic = False
        self.scPCA = 7
        self.scPCAMax = 12
        self.scPCAVariance = 99.96
        self.scAutoIters = True
        self.scMinImprov = 5
        self.sgfDo = False
        self.sgfWindow = 9
        self.sgfOrder = 3
        self.srDo = False
        self.srMin = 800
        self.srMax = 4000
        self.bcDo = False
        self.bcMethod = 'none'
        self.bcIters = 10
        self.bcLambda = 10000
        self.bcP = 0.01
        self.bcLambdaArpls = 10000
        self.bcThreshold = 0.01
        self.bcPoly = 3
        self.normDo = False
        self.normMethod = 'none'
        self.normWavenum = 1655

    def save(self, filename):
        with open(filename, 'w') as fp:
            json.dump(vars(self), fp, indent=4)

    def load(self, filename):
        with open(filename, 'r') as fp:
            data = json.load(fp)
            self.__dict__.update(data)



class PrepWorker(QObject):
    """
    Worker thread class for the heavy parts of the preprocessing: Scattering correction
    and the multiple-file batch processing.
    """
    # Signals for when processing is finished or failed, and for progress indication
    done = pyqtSignal(np.ndarray, np.ndarray, np.ndarray)
    stopped = pyqtSignal()
    failed = pyqtSignal(str, str)
    progress = pyqtSignal(int, int)
    progressPlot = pyqtSignal(np.ndarray, tuple)

    fileLoaded = pyqtSignal(int)
    loadFailed = pyqtSignal(str, str, str)

    batchProgress = pyqtSignal(int, int)
    batchDone = pyqtSignal(str)

    def __init__(self, parent=None):
        QObject.__init__(self, parent)
        self.halt = False

    @pyqtSlot(int)
    def loadFile(self, data, num):
        "Load file number num in the data object, emitting a signal on failure"
        try:
            file = data.filenames[num]
            if file == data.curFile:
                return True
            data.read_matrix(file)
        except (RuntimeError, FileNotFoundError) as e:
            self.loadFailed.emit(file, str(e), '')
        except Exception as e:
            self.loadFailed.emit(file, str(e), traceback.format_exc())
        else:
            self.fileLoaded.emit(num)
            return True
        return False

    def emitProgress(self, *pargs):
        "Combined progress signal and check for user interruption"
        if self.halt:
            raise InterruptedError('interrupted by user')
        self.progress.emit(*pargs)

    def loadReference(self, data, ref, otherref):
        """
        Load an RMieSC reference spectrum, at wavenumbers that match the currently
        loaded file.
        """
#        try:
        if ref.lower() == 'other':
            if otherref == '':
                raise RuntimeError('Specify a reference spectrum file')
            return correction.load_reference(data.wavenumber, matfilename=otherref)
        else:
            return correction.load_reference(data.wavenumber, what=ref.lower())
#        except FileNotFoundError as e:
#            raise RuntimeError('Unable to load reference spectrum: ' + str(e))

    def callACandSC(self, data, params, wn, y):
        """ Helper function for running RMieSC and/or atmospheric correction
        """
        if params.acDo:
            self.emitProgress(-1, 100)
            y = correction.atmospheric(wn, y, cut_co2=params.acSpline,
                                       extra_iters=5 if params.acLocal else 0,
                                       smooth_win=9 if params.acSmooth else 0,
                                       atm=params.acReference,
                                       progressCallback=self.emitProgress)[0]

        if params.scDo:
            self.emitProgress(-2, 100)
            ref = self.loadReference(data, params.scRef, params.scOtherRef)
            yold = y
            clust = params.scClusters * (-1 if params.scStable else 1) if params.scClustering else 0
            algos = {'Konevskikh': True, 'Bassan': False}
#            print(params.scAmin, params.scAmax, params.scResolution)
            y = correction.rmiesc(
                    wn, y, ref,
                    iterations=params.scIters,
                    clusters=clust,
                    n_components=params.scPCAMax if params.scPCADynamic else params.scPCA,
                    pcavariancelimit=params.scPCAVariance*.01 if params.scPCADynamic else 0,
                    a=np.linspace(params.scAmin, params.scAmax, params.scResolution),
                    d=np.linspace(params.scDmin, params.scDmax, params.scResolution),
                    bvals=params.scResolution,
                    konevskikh=algos[params.scAlgorithm],
                    linearcomponent=params.scLinear,
                    weighted=False,
                    autoiterations=params.scAutoIters,
                    targetrelresiduals=1-params.scMinImprov*.01,
                    progressCallback=self.emitProgress,
                    progressPlotCallback=self.progressPlot.emit,
                    verbose=True,
                    renormalize=params.scRenormalize)
            self.done.emit(wn, yold, y)
        return y

    @pyqtSlot(SpectralData, dict)
    def rmiesc(self, data, params):
        """ Run RMieSC, possibly preceded by atmospheric correction, on all or a subset of
        the raw data.
        Parameters:
            data: SpectralData object with raw data
            params: dictionary, mostly with things from PrepParameters (see code)
        """
        try:
            params['scDo'] = True
            params = namedtuple('rmiescParams', params.keys())(*params.values())
            if params.selected is not None:
                y = data.raw[params.selected]
            else:
                y = data.raw
            self.callACandSC(data, params, data.wavenumber, y)

        except InterruptedError:
            self.stopped.emit()
        except Exception as e:
            traceback.print_exc()
            self.failed.emit(repr(e), traceback.format_exc())

    @pyqtSlot(SpectralData, PrepParameters, str, bool)
    def bigBatch(self, data, params, folder, preservepath):
        """
        Run the batch processing of all the files listed in 'data'
        Parameters:
            data: SpectralData object with one or more files
            params: PrepParameters object from the user
            folder: output directory
            preservepath: if True, all processed files whose paths are under data.foldername
            will be placed in the corresponding subdirectory of the output directory.
        """
        try:
#            scipy.io.savemat(os.path.join(folder, '00.start'), {'nada': [[0]] } )

            for fi in range(len(data.filenames)):
                self.batchProgress.emit(fi, len(data.filenames))
                if not self.loadFile(data, fi):
                    continue

                wn = data.wavenumber
                y = data.raw

                y = self.callACandSC(data, params, wn, y)

                if params.sgfDo:
                    self.emitProgress(-3, 100)
                    y = scipy.signal.savgol_filter(y, params.sgfWindow, params.sgfOrder, axis=1)

                if params.srDo:
                    a = len(wn) - wn[::-1].searchsorted(params.srMax, 'right')
                    b = len(wn) - wn[::-1].searchsorted(params.srMin, 'left')
                    wn = wn[a:b]
                    y = y[:, a:b]

                if params.bcDo:
                    self.emitProgress(-4, 100)
                    if params.bcMethod == 'rubberband':
                        y -= baseline.rubberband(
                                wn, y, progressCallback=self.emitProgress)
                    elif params.bcMethod == 'concaverubberband':
                        y -= baseline.concaverubberband(
                                wn, y, iters=params.bcIters,
                                progressCallback=self.emitProgress)
                    elif params.bcMethod == 'asls':
                        y -= baseline.asls(
                                y, lam=params.bcLambda, p=params.bcP,
                                progressCallback=self.emitProgress)
                    elif params.bcMethod == 'arpls':
                        y -= baseline.arpls(
                                y, lam=params.bcLambdaArpls,
                                progressCallback=self.emitProgress)
                    else:
                        raise ValueError('unknown baseline correction method '+str(params.bcMethod))

                if params.normDo:
                    y = normalization.normalize_spectra(params.normMethod, y, wn,
                                                        wavenum=params.normWavenum)

                # Figure out where to save the file
                filename = data.curFile
                if preservepath and filename.startswith(data.foldername):
                    filename = filename[len(data.foldername):]
                    filename = folder + filename
                    os.makedirs(os.path.dirname(filename), exist_ok=True)
                else:
                    filename = os.path.join(folder, os.path.basename(filename))
                # Add the extension
                filename = os.path.splitext(filename)
                filename = filename[0] + params.saveExt + filename[1]

                y = np.hstack((wn[:,None], y.T))
                scipy.io.savemat(filename, {'AB': y, 'wh': data.wh } )

        except InterruptedError:
            self.stopped.emit()
        except Exception as e:
            traceback.print_exc()
            self.failed.emit(repr(e), traceback.format_exc())

        self.batchDone.emit('')


class ABCWorker(QObject):
    """
    A smaller worker thread class for atmospheric and baseline correction only.
    """
    acDone = pyqtSignal(np.ndarray, np.ndarray, np.ndarray, np.ndarray)
    acFailed = pyqtSignal(str)
    bcDone = pyqtSignal(np.ndarray, np.ndarray, np.ndarray)
    bcFailed = pyqtSignal(str)

    def __init__(self, parent=None):
        QObject.__init__(self, parent)
        self.haltBC = False

    @pyqtSlot(np.ndarray, np.ndarray, dict)
    def ac(self, wn, y, params):
        """
        Run baseline correction, emitting the processed data
        """
        try:
            corr, factors = correction.atmospheric(wn, y, cut_co2=params['cut_co2'],
                                          extra_iters=5 if params['extra'] else 0,
                                          smooth_win=9 if params['smooth'] else 0,
                                          atm=params['ref'])
            self.acDone.emit(wn, y, corr, factors)
        except Exception:
            self.acFailed.emit(traceback.format_exc())

    def checkHaltBC(self, a, b):
        if self.haltBC:
            raise InterruptedError('interrupted by user')

    @pyqtSlot(np.ndarray, np.ndarray, str, dict)
    def bc(self, wn, y, method, params):
        try:
            self.checkHaltBC(0, 1)
            if method in {'rubberband', 'concaverubberband'}:
                corr = getattr(baseline, method)(wn, y, **params,
                              progressCallback=self.checkHaltBC)
            else:
                corr = getattr(baseline, method)(y, **params,
                              progressCallback=self.checkHaltBC)
            self.bcDone.emit(wn, y, corr)
        except InterruptedError:
            self.bcFailed.emit('')
        except Exception:
            self.bcFailed.emit(traceback.format_exc())




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 28 01:07:18 2019

@author: carl
"""

from PyQt5.QtCore import pyqtSignal, pyqtSlot
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import numpy as np
import random
from sklearn.cluster import KMeans
from sklearn.metrics import pairwise_distances_argmin_min

class ProjectionWidget(FigureCanvas):
    changedSelected = pyqtSignal(int)

    def __init__(self, parent=None):
        self.wavenumber = None
        self.raw = None
        self.wh = None
        self.projection = np.random.rand(16,16)
        self.fig = Figure()
        self.ax = self.fig.add_subplot(111)
        self.ax.set_axis_off()
        self.ax.set_position([0,0,1,1])
        FigureCanvas.__init__(self, self.fig)
        self.cmap = 'hot'
        self.mpl_connect('button_press_event', self.onclick)
        # (x,y) -> Rectangle for the main plot
        self.selected = {}
        self.clustered = False
        self.popax = None
        # (x,y) -> Rectangle for pop-out, iff popax is not None
        self.popsel = {}
        self.mainimg = self.ax.imshow(self.projection, self.cmap, zorder=0, aspect='equal')
        self.popimg = None
        self.popcb = None
        self.popfig = None
        self.refreshplots = True

    def setData(self, wn, raw, wh):
#        if self.projection.shape != data.shape:
        if self.raw is None or self.raw.shape != raw.shape:
            self.clearSelected()
#            self.mainimg = None
#            self.popax = None
        self.wavenumber = wn
        self.raw = raw
        self.setDimensions(wh)

    def setDimensions(self, wh):
        p = wh[0] * wh[1]
        if p <= 0:
            return
        if wh[0] > 8 * wh[1] or wh[1] > 8 * wh[0]:
            h = int(np.sqrt(p))
            self.wh = ((p + h - 1) // h, h)
        else:
            self.wh = wh
        self.refreshplots = True
        self.draw_idle()

    def getWavenumbers(self):
        return self.wavenumber

    def getSelectedData(self):
        if self.raw is None:
            return None
        return self.raw[self.getSelected(), :]

    def setProjection(self, method, wavenumix):
        if self.wavenumber is None:
            return
        if method == 0:
            data = np.trapz(self.raw, self.wavenumber, axis=1)
            if self.wavenumber[0] > self.wavenumber[-1]:
                data = -data
        elif method == 1:
            data = self.raw.max(axis=1)
        elif method == 2:
            data = self.raw[:, wavenumix]
        px = self.wh[0] * self.wh[1]
        if px != len(data):
            d2 = np.empty((px))
            d2[:len(data)] = data
            d2[len(data):] = data.min()
            data = d2

        data = data.reshape(self.wh[::-1])
        if data.shape != self.projection.shape:
            sel = self.getSelected()
            self.projection = data
            self.setSelected(sel)
        else:
            self.projection = data
        self.refreshplots = True
        self.draw_idle()

    def clearSelected(self):
        for p in self.selected.values():
            p.remove()
        for p in self.popsel.values():
            p.remove()
        self.selected = {}
        self.popsel = {}

    def getSelected(self):
        return [xy[1] * self.projection.shape[1] + xy[0] for xy in self.selected.keys()]

    def removePoint(self, xy):
        self.selected.pop(xy).remove()
        if self.popax:
            self.popsel.pop(xy).remove()

    def newRect(self, xy):
        return patches.Rectangle((xy[0], xy[1]), 1, 1, fill=False, color='b')

    def addPoint(self, xy):
        px = xy[1] * self.projection.shape[1] + xy[0]
        if px >= len(self.raw) or px < 0:
            return False
        p = self.newRect(xy)
        self.ax.add_patch(p)
        self.selected[xy] = p
        if self.popax:
            p = self.newRect(xy)
            self.popax.add_patch(p)
            self.popsel[xy] = p
        return True

    def setSelected(self, sel):
        w = self.projection.shape[1]
        self.clearSelected()
        for xy in { (s % w, s // w) for s in sel }:
            self.addPoint(xy)
        self.draw_idle()
        self.changedSelected.emit(len(self.selected))

    def setSelectedCount(self, n, clustered):
        if self.raw is None:
            return
        if (n == len(self.selected) and clustered == self.clustered) and not self.refreshplots:
            return
        self.clustered = clustered
        if clustered and n > 0 and self.raw is not None:
            # (not wasc or n != len(self.selected)) and
            km = KMeans(n_clusters=min(n, len(self.raw)), n_init=1)
            km.fit(self.raw)
            closest, _ = pairwise_distances_argmin_min(km.cluster_centers_, self.raw)
            w = self.projection.shape[0]
            self.clearSelected()
            for xy in { (s % w, s // w) for s in closest }:
                self.addPoint(xy)
        if n < len(self.selected):
            for xy in list(self.selected.keys())[n:]:
                self.removePoint(xy)
        while len(self.selected) < n and len(self.selected) < len(self.raw):
            x = random.randrange(self.projection.shape[1])
            y = random.randrange(self.projection.shape[0])
            if (x, y) not in self.selected:
                self.addPoint((x, y))
        self.draw_idle()
        self.changedSelected.emit(len(self.selected))

    @pyqtSlot('QString')
    def setCmap(self, s):
        self.cmap = s
        if self.mainimg:
            self.mainimg.set_cmap(s)
        if self.popimg:
            self.popimg.set_cmap(s)
        self.draw_idle()

    def onclick(self, event):
        if not (event.inaxes == self.ax or event.inaxes == self.popax):
            return
        if event.xdata is None or self.wh is None:
            return
        hw = self.projection.shape
        x, y = (int(event.xdata), int(event.ydata))
        if not (0 <= x < hw[1] and 0 <= y < hw[0]):
            return
        if (x, y) in self.selected:
            self.removePoint((x, y))
        elif not self.addPoint((x, y)):
            return
        FigureCanvas.draw_idle(self)
        self.changedSelected.emit(len(self.selected))

    def draw(self):
        if self.refreshplots:
            self.mainimg.set_data(self.projection)
            self.mainimg.set_extent((0, self.projection.shape[1], self.projection.shape[0], 0))
            self.mainimg.autoscale()

        FigureCanvas.draw(self)
        if plt.fignum_exists(self.objectName()):
            self.popOut()
        self.refreshplots = False


    def popOut(self):
        setext = True
        fig = plt.figure(self.objectName(), tight_layout=dict(pad=.6))
        if self.popfig != fig:
            self.popfig = fig
            fig.canvas.mpl_connect('button_press_event', self.onclick)
            self.popax = fig.gca()
            self.popimg = self.popax.imshow(self.projection, self.cmap)
            self.popcb = fig.colorbar(self.popimg, ax=self.popax)
            setext = True

            self.popsel = {}
            for k in self.selected:
                p = self.newRect(k)
                self.popsel[k] = p
                self.popax.add_patch(p)
            fig.show()

        elif self.refreshplots:
            self.popimg.set_data(self.projection)
            self.popcb.mappable.set_clim(vmin=self.projection.min(), vmax=self.projection.max())
            setext = True

        if setext:
            self.popimg.set_extent((0, self.projection.shape[1], self.projection.shape[0], 0))
            self.popimg.autoscale()
        fig.canvas.draw_idle()




import os
import os.path
#import fnmatch
#import traceback
from pkg_resources import resource_filename
from io import BytesIO

#import numpy as np

from PyQt5.QtWidgets import QWidget
#from PyQt5.QtCore import QThread, pyqtSignal, pyqtSlot, Qt
from PyQt5 import uic

from . import constants, uitools

ImageVisualizerUi = uic.loadUiType(resource_filename(__name__, "imagevisualizer.ui"))[0]

class ImageVisualizerWidget(QWidget, ImageVisualizerUi):
    """
    Widget with a bunch of graphical components related to ...
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)
        self.lineEditWavenumber.setFormat("%.2f")

    def saveParameters(self, p):
        "Copy from UI to some kind of parameters object"
        p.plotMethod = self.comboBoxMethod.currentIndex()
        p.plotColors = self.comboBoxCmaps.currentText()
        p.plotWavenum = self.lineEditWavenumber.value()

    def loadParameters(self, p):
        "Copy to UI from some kind of parameters object"
        self.comboBoxMethod.setCurrentIndex(p.plotMethod)
        self.comboBoxCmaps.setCurrentText(p.plotColors)
        self.lineEditWavenumber.setValue(p.plotWavenum)


class ImageVisualizer():
    """
    Mixin class that adds hyperspectral image display to an OctavvsMainWindow.

    Things assumed to exist in self:
        imageVisualizer - ImageVisualizer (from UI)
        imageProjection
        plot_visual
        data

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        iv = self.imageVisualizer
        iv.comboBoxMethod.currentIndexChanged.connect(self.imageProjection)
        iv.horizontalSliderWavenumber.valueChanged.connect(self.wavenumberSlide)
        iv.lineEditWavenumber.editingFinished.connect(self.wavenumberEdit)
        iv.comboBoxCmaps.currentTextChanged.connect(self.plot_visual.setCmap)


    def updateWavenumberRange(self):
        super().updateWavenumberRange()
        # Update sliders before boxes to avoid errors from triggered updates
        iv = self.imageVisualizer
        if self.data.wavenumber is not None:
            iv.horizontalSliderWavenumber.setMaximum(len(self.data.wavenumber)-1)

        iv.labelMinwn.setText("%.2f" % self.data.wmin)
        iv.labelMaxwn.setText("%.2f" % self.data.wmax)
        wmin = min(self.data.wmin, constants.WMIN)
        wmax = max(self.data.wmax, constants.WMAX)
        iv.lineEditWavenumber.setRange(wmin, wmax, default=.5*(wmin+wmax))

        # Make sure the sliders are in sync with the boxes
        self.wavenumberEdit()

    def wavenumberEdit(self):
        iv = self.imageVisualizer
        uitools.box_to_slider(
                iv.horizontalSliderWavenumber, iv.lineEditWavenumber,
                self.plot_visual.getWavenumbers(),
                uitools.ixfinder_nearest)

    def wavenumberSlide(self):
        iv = self.imageVisualizer
        uitools.slider_to_box(
                iv.horizontalSliderWavenumber, iv.lineEditWavenumber,
                self.plot_visual.getWavenumbers(),
                uitools.ixfinder_nearest)
        self.imageProjection()

    def updateDimensions(self, wh):
        try:
            self.plot_visual.setDimensions(wh)
        except ValueError:
            pass
        self.imageProjection()

    def imageProjection(self):
        iv = self.imageVisualizer
        meth = iv.comboBoxMethod.currentIndex()
        iswn = meth == 2
        iv.lineEditWavenumber.setEnabled(iswn)
        iv.horizontalSliderWavenumber.setEnabled(iswn)
        self.plot_visual.setProjection(meth, -1 - iv.horizontalSliderWavenumber.value())

    # White light image stuff
    def loadWhite(self):
        filename = self.getImageFileName(title="Load white light image",
                                         settingname='whitelightDir')
        if filename:
            self.plot_whitelight.load(filename)

    def reloadWhiteLight(self):
        if self.data.image is not None:
            self.plot_whitelight.load(BytesIO(self.data.image[0]), self.data.image[1])
        else:
            self.plot_whitelight.load(os.path.splitext(self.data.curFile)[0]+'.jpg')


